unit MandelbrotCanvas;

interface

uses w3system, w3graphics, w3components, w3time,
     MandelbrotEngine,
     GpW3, GpCoordinateMapper, GpEmbelish;

const CMaxIterations = 50;//1000;

type
  TW3MandelbrotCanvas = class(TW3GraphicControl)
  private
    FCoord       : TCoordinateMapping;
    FEmbelish    : TEmbelishCanvas;
    FEngine      : TMandelbrotEngine;
    FInitialized : boolean;
    FIterations  : integer;
    FResumePaint : integer;
    FTimer       : TW3EventRepeater;
  protected
    procedure AdjustWorldBottom;
    function  CalcWorldBottom(left, top, right: float): float;
    procedure ClearView;
    procedure FinalizeObject; override;
    function  GetWorld: TFloatRect;
    procedure InitializeObject; override;
    procedure MakeInitialWorld;
    function  MapIterationToColor(iteration: integer): TW3RGBA;
    procedure MapIterationsToImageLine(const iterations: TIterationArray;
      imageLine: TW3ImageData);
    function  Normalize(rect: TRect): TRect;
    procedure Paint; override;
    function  PaintTimer(sender: TObject): boolean;
    procedure Resize; override;
    procedure SetIterations(value: integer);
    procedure SetWorld(const world: TFloatRect);
    function  UpDown(y: integer): integer;
  public
    function  GetImage: TW3ImageData;
    procedure HideOverlay;
    procedure Initialize;
    procedure MoveView(dX, dY: integer);
    procedure PutImage(image: TW3ImageData);
    procedure ShowArrow(pFrom, pTo: TPoint);
    procedure ShowMarquee(rect: TRect);
    procedure ZoomView(rect: TRect);
    function  ZoomLevel: float;
    property Iterations: integer read FIterations write SetIterations;// default CMaxIterations
    property World: TFloatRect read GetWorld write SetWorld;
  end;

implementation

const CInitialLeft = -0.8;
const CInitialTop = -1.2;
const CInitialRight = 2.2;
const CPaintTimeout_ms = 1000;

{ TW3MandelbrotCanvas }

procedure TW3MandelbrotCanvas.AdjustWorldBottom;
var
  botDelta: float;
begin
  botDelta := CalcWorldBottom(FCoord.World.Left, FCoord.World.Top, FCoord.World.Right) - FCoord.World.Bottom;
  FCoord.World := FloatRect(FCoord.World.Left, FCoord.World.Top - botDelta/2,
    FCoord.World.Right, FCoord.World.Bottom + botDelta/2);
end;

function TW3MandelbrotCanvas.CalcWorldBottom(left, top, right: float): float;
begin
  if FCoord.View.Right = FCoord.View.Left then
    Result := top
  else
    Result := top + FCoord.View.Height * (right - left) / FCoord.View.Width;
end;

procedure TW3MandelbrotCanvas.ClearView;
begin
  Canvas.StrokeStyle := '#000000';
  Canvas.LineWidth := 1;
  Canvas.FillRectF(FCoord.View.Left, FCoord.View.Top, FCoord.View.Width, FCoord.View.Height);
end;

procedure TW3MandelbrotCanvas.FinalizeObject;
begin
  FEmbelish.Free;
  FTimer.Free;
  FEngine.Free;
  FCoord.Free;
  inherited;
end;

function TW3MandelbrotCanvas.GetImage: TW3ImageData;
begin
  Result := Canvas.GetImageData(0, 0, FCoord.View.Width + 1, FCoord.View.Height + 1);
end;

function TW3MandelbrotCanvas.GetWorld: TFloatRect;
begin
  Result := FCoord.World;
end;

procedure TW3MandelbrotCanvas.HideOverlay;
begin
  FEmbelish.Hide;
end;

procedure TW3MandelbrotCanvas.Initialize;
begin
  MakeInitialWorld;
  Invalidate;
end;

procedure TW3MandelbrotCanvas.InitializeObject;
begin
  inherited;
  FIterations := CMaxIterations;
  FResumePaint := -1;
  FEngine := TMandelbrotEngine.Create;
  FEngine.MaxIterations := CMaxIterations;
  FCoord := TCoordinateMapping.Create;
  FEmbelish := TEmbelishCanvas.Create(Canvas);
end;

procedure TW3MandelbrotCanvas.MakeInitialWorld;
var
  offset, scale: float;
begin
  FCoord.World := FloatRect(CInitialLeft, CInitialTop, CInitialRight,
    CalcWorldBottom(CInitialLeft, CInitialTop, CInitialRight));
  // center vertically
  FCoord.World := FloatRect(FCoord.World.Left, - FCoord.World.Height / 2,
    FCoord.World.Right, FCoord.World.Height / 2);
  // scale if screen height is insufficient
  scale := 2 * Abs(CInitialTop) / FCoord.World.Height;
  if scale > 1 then
    FCoord.World.SetBounds(
      scale * FCoord.World.Left, - scale * FCoord.World.Height / 2,
      scale * FCoord.World.Right,  scale * FCoord.World.Height / 2);
  // move right if there is place
  if FCoord.World.Right > CInitialRight then begin
    offset := - (FCoord.World.Right - CInitialRight) / 2;
    FCoord.World.SetBounds(
      offset + FCoord.World.Left, - FCoord.World.Height / 2,
      offset + FCoord.World.Right,  FCoord.World.Height / 2);
  end;
  FInitialized := true;
end;

procedure TW3MandelbrotCanvas.MapIterationsToImageLine(
  const iterations: TIterationArray; imageLine: TW3ImageData);
var
  x: integer;
begin
  Assert(iterations.Length = imageLine.Width + 1);
  for x := 0 to iterations.Length - 1 do
    imageLine.setPixel(x, 0, MapIterationToColor(iterations[x]));
end;

function TW3MandelbrotCanvas.MapIterationToColor(iteration: integer): TW3RGBA;
var
  R, G, B: integer;
begin
  if iteration = FIterations then
    Result := MakeColor(0, 0, 0, 255)
  else begin
    iteration := iteration * 128;
    B := iteration mod 256;
    iteration := (iteration div 256) * 128;
    G := iteration mod 256;
    iteration := (iteration div 256) * 128;
    R := iteration mod 256;
    Result := MakeColor(R, G, B, 255);
  end;
end;

procedure TW3MandelbrotCanvas.MoveView(dX, dY: integer);
begin
  ClearView;
  FCoord.OffsetWorldByViewUnits(dX, - dY);
  FResumePaint := -1;
  Invalidate;
end;

function TW3MandelbrotCanvas.Normalize(rect: TRect): TRect;
var
  wHeight: float;
begin
  // Forces rectangle to have same proportions as the view by changing Bottom and
  // rearranges points so that Left and Top are really left and top.
  wHeight:= CalcWorldBottom(rect.Left, rect.Top, rect.Right) - rect.Top;
  rect.Bottom := rect.Top + (Abs(Round(wHeight)) * Sign(rect.Bottom - rect.Top));
  Result := FixOrientation(rect);
end;

procedure TW3MandelbrotCanvas.Paint;
var
  firstY    : integer;
  imageLine : TW3ImageData;
  iterations: TIterationArray;
  start     : float;
  y         : integer;
begin
  if not Visible then 
    Exit;
  if not FInitialized then
    MakeInitialWorld;
  imageLine := Canvas.CreateImageData(FCoord.View.Width + 1, 1);
  start := Now;
  firstY := FResumePaint;
  if firstY < 0 then begin
    ClearView;
    firstY := FCoord.View.Top;
  end;
  FResumePaint := -1;
  for y := firstY to FCoord.View.Bottom do begin
    iterations := FEngine.GetLine(FCoord.World.Left, FCoord.World.Right, FCoord.MapToWorldY(y), FCoord.View.Width + 1);
    MapIterationsToImageLine(iterations, imageLine);
    Canvas.PutImageData(imageLine, FCoord.View.Left, UpDown(y));
    if HasElapsedMs(start, CPaintTimeout_ms) and (y < FCoord.View.Bottom) then begin
      FResumePaint := y + 1;
      if not assigned(FTimer) then
        FTimer := TW3EventRepeater.Create(PaintTimer, 1);
      Exit;
    end;
  end;
  FTimer.Free;
  FTimer := nil;
end;

function TW3MandelbrotCanvas.PaintTimer(sender: TObject): boolean;
begin
  Paint;
  Result := false;
end;

procedure TW3MandelbrotCanvas.PutImage(image: TW3ImageData);
begin
  Canvas.PutImageData(image, 0, 0);
end;

procedure TW3MandelbrotCanvas.Resize;
var
  newView: TIntRect;
begin
  inherited;
  newView := IntRect(0, 0, Width - 1, Height - 1);
  FCoord.View := newView;
  if FInitialized then
    AdjustWorldBottom;
  FResumePaint := -1;
end;

procedure TW3MandelbrotCanvas.SetIterations(value: integer);
begin
  if value <> FIterations then begin
    FIterations := value;
    ClearView;
    FResumePaint := -1;
    FEngine.MaxIterations := FIterations;
    Invalidate;
  end;
end;

procedure TW3MandelbrotCanvas.SetWorld(const world: TFloatRect);
begin
  FCoord.World := world;
end;

procedure TW3MandelbrotCanvas.ShowArrow(pFrom, pTo: TPoint);
begin
  FEmbelish.ShowArrow(pFrom, pTo);
end;

procedure TW3MandelbrotCanvas.ShowMarquee(rect: TRect);
begin
  FEmbelish.ShowMarquee(Normalize(rect));
end;

function TW3MandelbrotCanvas.UpDown(y: integer): integer;
begin
  Result := FCoord.View.Top - y + FCoord.View.Bottom;
end;

procedure TW3MandelbrotCanvas.ZoomView(rect: TRect);
begin
  ClearView;
  rect.Top := UpDown(rect.Top);
  rect.Bottom := UpDown(rect.Bottom);
  rect := Normalize(rect);
  FCoord.SetWorldToViewUnits(TRect.Create(rect.Left, rect.Top, rect.Right, rect.Bottom));
  FResumePaint := -1;
  Invalidate;
end;

function TW3MandelbrotCanvas.ZoomLevel: float;
begin
  if FCoord.World.Width = 0 then
    Result := 1
  else
    Result := (CInitialRight - CInitialLeft) / FCoord.World.Width;
end;
